/*****************************************************************************
 * $RCSfile: AttackBehaviour.cc,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 *
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: seekamp $
 * $Date: 2002/04/10 13:24:53 $
 * $Revision: 1.20 $
 *
 *****************************************************************************/

#include "kick.h"
#include "AttackBehaviour.hh"
#include "BTeamAction.hh"
#include "client.h"
#include "Log.hh"
#include "mybehave.h"

// Radius um Gegner; Mindestwert
#define OPP_RAD_MIN 1.5
// Radius um Gegner; vergroessert sich um so viel m pro m Abstand
#define OPP_RAD_FAC 0.3

float AttackBehaviour::calcRadiusForOpponent( cmu::PlayerObject * tempPlayerObj ) {
  //
  float dist = cmu::Mem->MyPos().dist( tempPlayerObj->get_abs_pos() );
  return OPP_RAD_MIN + ( dist * ( OPP_RAD_FAC ) ); // mindestwert + abhaengig von abstand einen faktor draufrechnen

}

inline float AttackBehaviour::distance(const cmu::Vector & a,const cmu::Vector & b)
{
  return sqrt(((b.x-a.x)*(b.x-a.x))+((b.y-a.y)*(b.y-a.y)));
}

inline float AttackBehaviour::maximum(float a, float b)
{
  if (a>b)
    return a;
  
  else
    return b;
}


/** Erzeugt eine Liste von Gegnern, die sich im Bereich einer moeglichen Abwehrkette befinden. */
list<cmu::PlayerObject *> * AttackBehaviour::createListOfDefendingOpponents() {
  //
  bool last = true;
  list<cmu::PlayerObject *> * defendingOpponents  = new list<cmu::PlayerObject *>;
  cmu::PlayerObject * tempPlayer = NULL;
  cmu::PlayerObject * tempPlayer2 = NULL;
  
  for (int i=0; i<cmu::Mem->NumOpponents();i++) 
    {
      tempPlayer = cmu::Mem->Opponents()[i];
      
      if ((tempPlayer != NULL) &&
	(tempPlayer->pos_valid())) {
	if (tempPlayer->get_abs_pos().x > cmu::Mem->BallX())
	  {
	   tempPlayer2 = *defendingOpponents->begin();
	   list<cmu::PlayerObject*>::iterator endIterator = defendingOpponents->end();
	   
	   for (list<cmu::PlayerObject*>::iterator iter = defendingOpponents->begin(); iter!=endIterator; iter++)
	     {
	       tempPlayer2 = *iter;
	       if (tempPlayer->get_abs_pos().y < tempPlayer2->get_abs_pos().y)  
		 {
		   defendingOpponents->insert( iter, tempPlayer );
		   last = false;
		   break;
		 }
	     }

	   if (last)
	     defendingOpponents->push_back(tempPlayer);
	   
	  }
      }
    }
  return defendingOpponents;
}

list<ComparableRay> * AttackBehaviour::getPossibleShotRays( const cmu::Vector & origin ) {
  //
  list<ComparableRay> * rays = new list<ComparableRay>;
  list<cmu::PlayerObject *> * players = createListOfDefendingOpponents();
  list<cmu::PlayerObject *>::iterator endIterator = players->end();

  for (list<cmu::PlayerObject *>::iterator iter = players->begin(); iter!=endIterator; iter++)
    {

      cmu::PlayerObject * tempPlayerObj = *iter;
      float radius = calcRadiusForOpponent( tempPlayerObj );
      cmu::Vector vec = tempPlayerObj->get_abs_pos();

      rays->push_back( leftTangentOfCircleThroughPoint(vec, radius, origin ) );
      rays->push_back( rightTangentOfCircleThroughPoint(vec, radius, origin ) );
    }

  delete players;

  return rays;
}

bool AttackBehaviour::teamBallPossession() {
  return (cmu::Mem->BallPossessor() > 0);
}

bool AttackBehaviour::behave() {

  Log::debug << teamBallPossession() << endl;
  if (cmu::Mem->PlayMode!=cmu::PM_Play_On)
    {
      onTheWay = 0;
      return false;
    }

   bool kickable = cmu::Mem->BallKickable();  
  if ( kickable && (chanceFuerSteilpass() )) {

    // Situation erkennen
    Log::debug << "Ballbesitz : " <<   kickable  << endl;
    Log::debug << "Meine Position " << cmu::Mem->MyX() << " " << cmu::Mem->MyY() ;
    Log::debug << "Ich will einen Steilpass spielen !" << endl;
    
    list<ComparableRay> * teamRays = getTeammateMoveRays();
    list<ComparableRay> * shotRays = getPossibleShotRays( cmu::Mem->BallAbsolutePosition() );

    bestPoint = getRayIntersects(teamRays,shotRays);
    Log::debug << "Ich passe zu :" << bestPoint << endl;
    
    /*
      An dieser Stelle gilt zu ueberlegen ob der bestPoint nicht zu weit in der 
      Naehe des geg. Torwarts liegt.
     */


    // ===== calculate distances to goalie

    float goalieTargetDist;
    float goalieMyselfDist;

    cmu::PlayerObject * goalie = cmu::Mem->GetPlayer( cmu::Mem->TheirSide, 1 );
    if ( goalie == 0 ) {
      // don't see goalie? guess distances!
      goalieTargetDist = 45-bestPoint.y; // Abschaetzung des Abstands basierend auf den y Koordinaten
      goalieMyselfDist = 45-cmu::Mem->MyY();

    } else {

      cmu::Vector g = goalie->get_abs_pos();
      goalieTargetDist = distance( bestPoint, g );
      goalieMyselfDist = distance( cmu::Mem->MyPos(), g );

    }

    float distThreshold = goalieMyselfDist * 3.0 / 5.0;
    
    if ( goalieTargetDist < distThreshold ) {
      Log::debug << "schnittpunkt zu weit vorne. besser nicht so weit schiessen." << endl;
      Log::debug << "alter Punkt: " << bestPoint << endl;
      bestPoint = ( ( bestPoint-cmu::Mem->MyPos() ).SetLength( distThreshold ) ) + cmu::Mem->MyPos();
      Log::debug << "neuer Punkt: " << bestPoint << endl;
    } 

    bestPoint.x = 0.7*bestPoint.x;
    bestPoint.y = 0.7*bestPoint.y;
    if (bestPoint.x < cmu::Mem->MyX())
      return false;
    
    cmu::Mem->doKick( Action( bestPoint ) );
    
    // smart_pass(bestPoint, cmu::KM_HardestKick);

    
    delete teamRays;
    delete shotRays;
    
    // Schnittpunkte suchen blabla...
    return true;
  } 
  // Log::debug << teamBallPossession() << endl;
  else if ((onTheWay>0) && (onTheWay<10) && (!cmu::Mem->OffsidePosition(cmu::Mem->MyPos(),cmu::Mem->MySide)))
    {    
      

      Log::debug << "on the way" << endl;
      if (opponentInterception())
	{
	  Log::debug << "Der Ball wurde abgefangen " << endl;
	  onTheWay = 0;
	  return false;
	}

      Log::debug << "Distanz MyPos,bestPoint" << distance(cmu::Mem->MyPos(),bestPoint) << endl;
      if (distance(cmu::Mem->MyPos(),bestPoint)<5.0)
	{
	  Log::debug << "Ich bin fast da, neu orientieren!" << endl;
	  onTheWay = 0;  
	  ballX3CyclesAgo = ballX2CyclesAgo = ballX1CycleAgo = ballX = -30.0;
	  return false;
	}
      
      Log::debug << "Bester Punkt : " << bestPoint << endl;
      cmu::go_to_point(bestPoint,1.5+fabs(cmu::Mem->BallX()-cmu::Mem->MyX())/10.0,100,cmu::DT_none);
      return true;      
    }

  // Log::debug << teamBallPossession() << endl;
  
  else if ( ( teamBallPossession() ) && (cmu::Mem->MyNumber>6) && (chanceToGetSteilpass()) && (onTheWay==0) && (!cmu::Mem->OffsidePosition(cmu::Mem->MyPos(),cmu::Mem->MySide)))
    {
      Log::debug<<"Steilpass - ich lauf los"<<endl;
      bestPoint = cmu::Vector(0.6 * ORIGINAL_FIELD_SIZE_X,getMySteilpassLaufwegY());

      if (distance(cmu::Mem->MyPos(),bestPoint)<5.0)
	return false;
      
      Log::debug << "Bester Punkt : " << bestPoint << endl;
      cmu::go_to_point(bestPoint,1.5+fabs(cmu::Mem->BallX()-cmu::Mem->MyX())/10.0,100,cmu::DT_none);
      onTheWay++;
      return true;
    }
  
  else return false;
  // ================================================================
  // test code follows here
}



bool AttackBehaviour::opponentInterception()
{
  ballX3CyclesAgo = ballX2CyclesAgo;
  ballX2CyclesAgo = ballX1CycleAgo;
  ballX1CycleAgo = ballX;
  ballX = cmu::Mem->BallX();
  // wenn sich der Ball ueber 3 Zyklen Richtung unser Tor bewegt hat, schliessen wir daraus, dass der Gegner ihn 
  // hat.

  if (ballX<=ballX3CyclesAgo)
    {
      ballX3CyclesAgo = ballX2CyclesAgo = ballX1CycleAgo = ballX = -30.0;
      return true;
    }  

  else return false;
}




float AttackBehaviour::getTeammateEvaluation(const cmu::Vector & origin,const cmu::Vector & target)
{
  list<cmu::PlayerObject *> * defendingOpponents  = createListOfDefendingOpponents();
  cmu::PlayerObject * tempPlayer = *defendingOpponents->begin();
  cmu::PlayerObject * oldPlayer = NULL;
  float firstDistance = 0;
  float secondDistance = 0;
  
  
  list<cmu::PlayerObject*>::iterator endIterator = defendingOpponents->end();
  for (list<cmu::PlayerObject*>::iterator iter = defendingOpponents->begin(); iter!=endIterator; iter++)
    {
      oldPlayer = tempPlayer;
      tempPlayer = *iter;
      if (origin.y <= tempPlayer->get_abs_pos().y)
	{
	  break;
	}
    }
  
  firstDistance = distance(tempPlayer->get_abs_pos(),target);
  if (oldPlayer!=NULL)
    secondDistance = distance(oldPlayer->get_abs_pos(),target);
  
  
  return (distance(origin,target)-max(firstDistance,secondDistance)-(((target.y>50)||(target.y<-50))?5:0));


}




float AttackBehaviour::lastOpponent()
{
  
  float goalie = 0.0;
  float lastOpponent = 0.0;
  cmu::PlayerObject * tempPlayer;
  
  
  for (int i=0; i<cmu::Mem->NumOpponents();i++) 
    {
      tempPlayer = cmu::Mem->Opponents()[i];
      
      if ((tempPlayer != NULL) &&
	  (tempPlayer->pos_valid())) {
	if (tempPlayer->get_abs_pos().x > cmu::Mem->BallX())
	  {
	    if (tempPlayer->get_abs_pos().x > goalie)
	      {
		lastOpponent = goalie;
		goalie = tempPlayer->get_abs_pos().x;
	      }
	    
	    else if ((tempPlayer->get_abs_pos().x>lastOpponent))
	      lastOpponent = tempPlayer->get_abs_pos().x;
	    
	  }     
      }
    }



  return lastOpponent;
}



bool AttackBehaviour::chanceFuerSteilpass()
{

  // Log::debug << "Last Opponent" << lastOpponent() << endl;
  // Log::debug << "Ballposition : " << cmu::Mem->BallAbsolutePosition() << endl;
  
  if ((lastOpponent()<(int)(0.5*ORIGINAL_FIELD_SIZE_X)) && (cmu::Mem->BallX()>(int)(-0.1*ORIGINAL_FIELD_SIZE_X)))
    return true;
  
  else
    return false;
}

bool AttackBehaviour::chanceToGetSteilpass()
{  
  // Log::debug << "TeammateChanceFuerSteilpass start" << endl;
        
  if ((lastOpponent()<(int)(0.5*ORIGINAL_FIELD_SIZE_X)) && (cmu::Mem->BallX()>(int)(-0.1*ORIGINAL_FIELD_SIZE_X)) && (cmu::Mem->MyX()>(-0.15*ORIGINAL_FIELD_SIZE_X)) && (cmu::Mem->MyX()<(0.2*ORIGINAL_FIELD_SIZE_X)))
    {
      Log::debug << "Chance to get Steilpass" << endl;
      return true;
    }
  
  else
    return false;
  

}
  

bool AttackBehaviour::TeammateChanceFuerSteilpass(cmu::PlayerObject * player)
{
  if (player == NULL)
    {
      Log::debug << "player == NULL" << endl;
      return false;
    }
  
  
  
  if ((lastOpponent()<(int)(0.5*ORIGINAL_FIELD_SIZE_X)) && (cmu::Mem->BallX()>(int)(-0.1*ORIGINAL_FIELD_SIZE_X)) && (player->get_abs_pos().x>(-0.15*ORIGINAL_FIELD_SIZE_X)))
    {
      return true;
    }
  
  else
    return false;
}






list<ComparableRay>* AttackBehaviour::getTeammateMoveRays()
{
  
  list<ComparableRay>* rays = new list<ComparableRay>;
  
  for (int i=0; i<cmu::Mem->NumTeammates();i++) 
    {
      cmu::PlayerObject * tempPlayer = cmu::Mem->Teammates()[i];
      
      if ((tempPlayer != NULL) &&
	  (tempPlayer->pos_valid())) 
	{
	  if (TeammateChanceFuerSteilpass(tempPlayer))
	    {
	      // Log::debug << "Teammate an Pos : " << tempPlayer->get_abs_pos() << "ist anspielbar" << endl;
	      ComparableRay ray = createRay(tempPlayer->get_abs_pos(),cmu::Vector(ORIGINAL_FIELD_SIZE_X,getSteilpassLaufwegY(tempPlayer)));
	      rays->push_back(ray);
	    }
	}
    }
  ComparableRay ray = createRay(cmu::Vector(0.3*ORIGINAL_FIELD_SIZE_X,-ORIGINAL_FIELD_SIZE_Y),cmu::Vector(0.3*ORIGINAL_FIELD_SIZE_X,ORIGINAL_FIELD_SIZE_Y));
  rays->push_back(ray);
  return rays;
}


double AttackBehaviour::getSteilpassLaufwegY(cmu::PlayerObject * player)
{

  double first_y=0;
  double second_y=0;
    
  list<cmu::PlayerObject *> * defendingOpponents  = createListOfDefendingOpponents();
  cmu::PlayerObject * tempPlayer = *defendingOpponents->begin();
  cmu::PlayerObject * oldPlayer = NULL;
  
  list<cmu::PlayerObject*>::iterator endIterator = defendingOpponents->end();
  for (list<cmu::PlayerObject*>::iterator iter = defendingOpponents->begin(); iter!=endIterator; iter++)
    {
      oldPlayer = tempPlayer;
      tempPlayer = *iter;
      // Log::debug << "y-Koordinate der Spieler in der Abwehrkette :" << tempPlayer->get_abs_pos().y << endl;
      if (player->get_abs_pos().y <= tempPlayer->get_abs_pos().y)  
	{
	  first_y = oldPlayer->get_abs_pos().y;
	  second_y = tempPlayer->get_abs_pos().y;
	  break;
	}
      if (first_y==second_y)
	first_y=first_y-(first_y+34.0)/2;
      
      if (second_y==0.0)
	second_y=first_y+(34.0-first_y)/2;
      
    }
  
  return (first_y+second_y)/2;
}

double AttackBehaviour::getMySteilpassLaufwegY()
{
  double first_y=0;
  double second_y=0;
  
  list<cmu::PlayerObject *> * defendingOpponents  = createListOfDefendingOpponents();
  cmu::PlayerObject * tempPlayer = *defendingOpponents->begin();
  cmu::PlayerObject * oldPlayer = NULL;
  
  list<cmu::PlayerObject*>::iterator endIterator = defendingOpponents->end();
  for (list<cmu::PlayerObject*>::iterator iter = defendingOpponents->begin(); iter!=endIterator; iter++)
    {
      oldPlayer = tempPlayer;
      tempPlayer = *iter;
      // Log::debug << "y-Koordinate der Spieler in der Abwehrkette :" << tempPlayer->get_abs_pos().y << endl;
      if (cmu::Mem->MyY() <= tempPlayer->get_abs_pos().y)  
	{
	  first_y = oldPlayer->get_abs_pos().y;
	  second_y = tempPlayer->get_abs_pos().y;
	  break;
	}
      if (first_y==second_y)
	first_y=first_y-(first_y+34.0)/2;
      
      if (second_y==0.0)
	second_y=first_y+(34.0-first_y)/2;
      
    }
  
  return (first_y+second_y)/2;
  
}







ComparableRay AttackBehaviour::createRay(const cmu::Vector & a,const cmu::Vector & target)
{
  cmu::Vector t = target;
  cmu::Vector b = a;
  
  cmu::AngleDeg   angleToTarget     = (t - b).dir();
  ComparableRay returnRay = ComparableRay( b, angleToTarget );
  return returnRay;
}






cmu::Vector AttackBehaviour::getRayIntersects(list<ComparableRay> * teamRays,list<ComparableRay> * shotRays)
{
  list<ComparableRay>::iterator endIteratorTeam = teamRays->end();
  list<ComparableRay>::iterator endIteratorShot = shotRays->end();
  cmu::Vector intersection;
  cmu::Vector bestPoint;
  double eval=0;
  double bestEval=0;
  

  for (list<ComparableRay>::iterator iterTeam = teamRays->begin(); iterTeam!=endIteratorTeam; iterTeam++)
    {
      ComparableRay teamRay = *iterTeam;
      for (list<ComparableRay>::iterator iterShot = shotRays->begin(); iterShot!=endIteratorShot; iterShot++)
	{
	  ComparableRay shotRay = *iterShot;
	  if (teamRay.intersection(shotRay,&intersection) )
	  {
	    eval = getTeammateEvaluation(teamRay.getOrigin(),intersection);  
	    if (eval<bestEval)
	      {
		bestEval = eval;
		bestPoint = intersection;
	      }
	  }
	}
    }
  
  return bestPoint;
  
}


// ================================================================
// test code follows here


bool AttackBehaviour::testLeftTangentOfCircleThroughPoint() {

  AttackBehaviour behaviour;

  cmu::Vector center(10,0);
  cmu::Vector point(0,0);
  float radius=2;

  ComparableRay erg = behaviour.leftTangentOfCircleThroughPoint( center, radius, point );
  ComparableRay erg2 = ComparableRay( cmu::Vector( 0,0 ), cmu::Vector( 10,-2 ) );

  return erg.similarto(erg2);
}
bool AttackBehaviour::testRightTangentOfCircleThroughPoint() {

  AttackBehaviour behaviour;

  cmu::Vector center(10,0);
  cmu::Vector point(0,0);
  float radius=2;

  ComparableRay erg = behaviour.rightTangentOfCircleThroughPoint( center, radius, point );
  ComparableRay erg2 = ComparableRay( cmu::Vector( 0,0 ), cmu::Vector( 10,2 ) );

  bool res = erg.similarto(erg2);
  if (!res) {
    erg.print();
    erg2.print();
  }

  return res;
}

/** Main test method for the AttackBehaviour component */
int AttackBehaviour::test() {
  int c = 0;
  if (!testLeftTangentOfCircleThroughPoint()) {
    cerr << "AttackBehaviour::testLeftTangentOfCircleThroughPoint() failed!" << endl;
    c++;
  }
  if (!testRightTangentOfCircleThroughPoint()) {
    cerr << "AttackBehaviour::testRightTangentOfCircleThroughPoint() failed!" << endl;
    c++;
  }
  
  return c;
}


/*
 * $Log: AttackBehaviour.cc,v $
 * Revision 1.20  2002/04/10 13:24:53  seekamp
 * Dribbling/Schuss verbessert.
 * Steilpaesse werden weniger hart gespielt.
 * Wenn aussen wird Richtung, Eckfahne gedribbelt
 *
 * Revision 1.19  2002/04/09 17:07:54  seekamp
 * kleine Korrekturen
 *
 * Revision 1.18  2002/04/09 12:28:58  seekamp
 * Dribbling Richtung Torwart, wenn ClosestPlayerToTheirGoal
 *
 * Revision 1.17  2002/04/08 17:59:47  fastjack
 * ScoringBehviour reimplemented. Does not work, yet. So don't worry if we don't shoot a goal...
 *
 * Revision 1.16  2002/04/08 15:49:54  seekamp
 * kleinere Verbesserungen
 *
 * Revision 1.15  2002/04/08 13:29:44  fastjack
 * & gegen && ersetzt
 *
 * Revision 1.14  2002/04/08 13:28:05  seekamp
 * Aktionsausfuehrung fuer Manndecker wieder Staminschonender implementiert
 *
 * Revision 1.13  2002/04/05 16:14:18  fastjack
 * zu lange Steilpaesse verkuerzt, kleinere Optimierungen
 *
 * Revision 1.12  2002/04/05 14:56:55  seekamp
 * In AttackBehaviour das Laufen zum Zielpunkt eines Steilpasses zielstrebiger gestaltet.
 *
 * Revision 1.11  2002/04/04 15:54:07  seekamp
 * kleinere Veraenderungen am Attack und Shot Behaviour zudem Libero wieder eingefuehrt und Aussenverteidiger bei gegnerischen Angriffen ueber die gegenueberliegende Seite Richtung Mitte gezogen.
 *
 * Revision 1.10  2002/04/02 15:52:49  seekamp
 * Positionsfindung der Abwehrspieler besser auf Abseitsfalle und Viererkette eingestellt.
 *
 * Revision 1.9  2002/04/01 15:57:11  fastjack
 * world.update wird nicht mehr im PotentialField aufgerufen. Grund: mehr Flexibilitaet. AttackBehaviour und BTeamAction von World entkoppelt.
 *
 * Revision 1.8  2002/03/28 16:19:45  seekamp
 * Anderes Verhalten vor Standardsituationen: es wird zunaechst 10 Zyklen gewartet, dann laeuft der ballnaechste Spieler zum Ball und schiesst direkt.
 *
 * Revision 1.7  2002/03/28 12:55:23  seekamp
 * Schnittpunkt fuer Steilpaesse erweitert. ( Passgeber kann sich den Ball auch selber vorlegen).
 *
 * Revision 1.6  2002/03/26 17:32:10  seekamp
 * Es muessten jetzt Steilpaesse gespielt werden, konnte allerdings noch nicht getestet werden.
 *
 * Revision 1.5  2002/03/26 00:33:01  fastjack
 * tests OK
 *
 * Revision 1.4  2002/03/26 00:17:54  fastjack
 * test code optimized
 *
 * Revision 1.3  2002/03/25 23:50:24  fastjack
 * bug fix
 *
 * Revision 1.2  2002/03/25 18:17:03  fastjack
 * Attack hat noch BUGS :o)
 *
 * Revision 1.1  2002/03/25 16:14:06  fastjack
 * Merged & added Behaviour classes for offensive behaviour
 *
 */




