/*****************************************************************************
 * $RCSfile: AttackBehaviour.hh,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 *
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: seekamp $
 * $Date: 2002/04/08 15:49:54 $
 * $Revision: 1.10 $
 *
 * TODO: move verbessern
 *
 *****************************************************************************/

#ifndef ATTACKBEHAVIOUR_H
#define ATTACKBEHAVIOUR_H

#include "Behaviour.hh"

namespace cmu
{
  class PlayerObject;
}

class BTeamAction;

class AttackBehaviour : public Behaviour {

  BTeamAction * actionController;

  float calcRadiusForOpponent( cmu::PlayerObject * tempPlayerObj );
  list<cmu::PlayerObject *> * createListOfDefendingOpponents();
  list<ComparableRay> * getPossibleShotRays( const cmu::Vector & origin );

  inline float distance(const cmu::Vector & a,const cmu::Vector & b);
  inline float maximum(float a, float b);
  float lastOpponent();
  bool chanceFuerSteilpass();
  list<ComparableRay>* getTeammateMoveRays();
  double getSteilpassLaufwegY(cmu::PlayerObject * player);
  double getMySteilpassLaufwegY();
  bool TeammateChanceFuerSteilpass(cmu::PlayerObject *  player);
  bool chanceToGetSteilpass();
  ComparableRay createRay(const cmu::Vector & a,const cmu::Vector & target);
  float getTeammateEvaluation(const cmu::Vector & origin,const cmu::Vector & target);
  cmu::Vector getRayIntersects(list<ComparableRay> * teamRays,list<ComparableRay> * shotRays);
  bool teamBallPossession();
  bool opponentInterception();
  
  
  // 
  private:  
  cmu::Vector bestPoint;
  int onTheWay;
  double ballX,ballX1CycleAgo,ballX2CyclesAgo,ballX3CyclesAgo;


  public:
  static const float ORIGINAL_FIELD_SIZE_X = 52.5;
  static const float ORIGINAL_FIELD_SIZE_Y = 34.0;
  
  AttackBehaviour( ) : actionController( 0 ),onTheWay(false),ballX1CycleAgo(-30.0),ballX2CyclesAgo(-30.0),ballX3CyclesAgo(-30.0) {};
  AttackBehaviour( BTeamAction * aC ) : actionController( aC ) {};
  virtual bool behave();
  
  // ================================================================
  // test code follows here
  
  static bool testLeftTangentOfCircleThroughPoint();
  static bool testRightTangentOfCircleThroughPoint();
  static int test();
};

#endif

/*
 * $Log: AttackBehaviour.hh,v $
 * Revision 1.10  2002/04/08 15:49:54  seekamp
 * kleinere Verbesserungen
 *
 * Revision 1.9  2002/04/05 16:14:18  fastjack
 * zu lange Steilpaesse verkuerzt, kleinere Optimierungen
 *
 * Revision 1.8  2002/04/05 14:56:55  seekamp
 * In AttackBehaviour das Laufen zum Zielpunkt eines Steilpasses zielstrebiger gestaltet.
 *
 * Revision 1.7  2002/04/02 15:52:49  seekamp
 * Positionsfindung der Abwehrspieler besser auf Abseitsfalle und Viererkette eingestellt.
 *
 * Revision 1.6  2002/04/01 15:57:11  fastjack
 * world.update wird nicht mehr im PotentialField aufgerufen. Grund: mehr Flexibilitaet. AttackBehaviour und BTeamAction von World entkoppelt.
 *
 * Revision 1.5  2002/03/28 12:55:23  seekamp
 * Schnittpunkt fuer Steilpaesse erweitert. ( Passgeber kann sich den Ball auch selber vorlegen).
 *
 * Revision 1.4  2002/03/26 17:32:11  seekamp
 * Es muessten jetzt Steilpaesse gespielt werden, konnte allerdings noch nicht getestet werden.
 *
 * Revision 1.3  2002/03/26 00:17:54  fastjack
 * test code optimized
 *
 * Revision 1.2  2002/03/25 18:17:03  fastjack
 * Attack hat noch BUGS :o)
 *
 * Revision 1.1  2002/03/25 16:14:06  fastjack
 * Merged & added Behaviour classes for offensive behaviour
 *
 */
