/*****************************************************************************
 * $RCSfile: CommunicationManager.cc,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *		       Jens Meyer, Robert Adolph
 * 
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: fastjack $
 * $Date: 2002/04/15 03:24:20 $
 * $Revision: 1.18 $
 *****************************************************************************/


#include "CommunicationManager.hh"
#include "BallMessage.hh"
#include "BTeamAction.hh"
#include "client.h"
#include <cstdlib>

void CommunicationManager::broadcast()	 // Communicate World State
{
  float vconf = cmu::Mem->BallVelocityValid();
  float conf = cmu::Mem->BallPositionValid();

  if ((vconf > 0.7)&&(conf > 0.7)) {

      float x = cmu::Mem->BallX();
      float y = cmu::Mem->BallY();
      float vx = cmu::Mem->GetBall()->get_abs_vel().x;
      float vy = cmu::Mem->GetBall()->get_abs_vel().y;
      float dist = cmu::Mem->BallDistance();

      BallMessage message( x,y,conf,vx,vy,vconf,dist );

      cmu::Mem->doSay(message.getCharString());
      if (debug)
	cout << "said: " << message.getCharString() << endl;
    }
}

void CommunicationManager::parse(char *message)
{
  if ( message[1] == BallMessage::IDENTIFIER ) {
    BallMessage * msg = new BallMessage( &message[1] );

    if (debug)
      cout << "heard: " << message << endl;

    cmu::Vector hearedPos( msg->getX(),msg->getY() );

    if ( hearedPos.dist( cmu::Mem->MyPos() ) > 6 )

	cmu::Mem->HearBall( msg->getX(), msg->getY(), msg->getConf(), msg->getVelX(), msg->getVelY(), msg->getVelConf(), msg->getDist(), cmu::Mem->CurrentTime );
  }
}

/*
 * $Log: CommunicationManager.cc,v $
 * Revision 1.18  2002/04/15 03:24:20  fastjack
 * BallMessage Klasse fr das Parsen/Generieren von BallInfos gebaut. CommunicationManager aufgerumt.
 *
 * Revision 1.17  2002/04/13 07:59:16  jens
 * German Open 2002 Edition
 *
 * Revision 1.16  2002/04/04 21:15:43  fastjack
 * removed PotentialField code from BTeamAction, all is in DefaultBehaviour now. I hope it works.
 *
 * Revision 1.15  2002/04/02 15:58:11  fastjack
 * Various changes. Torwart Abschlag Bugfix. kleinere unbedeutendere Korrekturen.
 *
 * Revision 1.14  2002/04/01 15:57:12  fastjack
 * world.update wird nicht mehr im PotentialField aufgerufen. Grund: mehr Flexibilitaet. AttackBehaviour und BTeamAction von World entkoppelt.
 *
 * Revision 1.13  2002/03/28 15:49:46  fastjack
 * Ball rumgeschreie vom FCP Agent eingebaut. Ausgaben wieder ausgebaut
 *
 * Revision 1.12  2002/03/28 12:52:11  fastjack
 * Absturzursache eingegrenzt - Player2Player Kommunikation entfernt als schnelle Loesung
 *
 * Revision 1.11  2002/03/26 17:56:22  fastjack
 * Performance Optimierungen.
 *
 * Revision 1.10  2002/03/25 16:14:06  fastjack
 * Merged & added Behaviour classes for offensive behaviour
 *
 * Revision 1.9  2002/03/15 13:26:00  fastjack
 * Ausgaben entfernt / dem Gouraud Kerl mehr Genauigkeit verschafft und einen boesen BUG entfernt
 *
 * Revision 1.8  2002/03/01 11:18:45  fastjack
 * GPL added
 *
 * Revision 1.7  2002/02/28 16:13:58  fastjack
 * bug fixed: Sporadische Abstuerze
 *
 * Revision 1.6  2002/02/28 11:30:52  fastjack
 * Ausgaben entfernt
 *
 * Revision 1.5  2002/02/21 17:24:58  seekamp
 * cone-Groesse veraendert
 *
 * Revision 1.4  2002/02/21 12:10:42  fastjack
 * Absturzproblem in CommunicationManager::send_game_status behoben
 *
 * Revision 1.3  2002/02/15 13:51:09  fastjack
 * GouraudFiller repariert, fehlerhaften Algorithmus gefixt. Paar Kommentare entfernt.
 *
 * Revision 1.2  2002/02/14 19:39:53  fastjack
 * Gouraud Shader verbessert - sind noch Probleme drin leider.
 *
 * Revision 1.1  2002/02/10 10:58:18  fastjack
 * Habe den Code, der sich mit der Kommunikation befasst, in eine neue Klasse
 * CommunicationManager herausgezogen, da die BTeamAction sonst zu gross und
 * unbersichtlich wird.
 *
 */

