/*****************************************************************************
 * $RCSfile: DefaultBehaviour.cc,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *		       Jens Meyer, Robert Adolph
 *
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: jens $
 * $Date: 2002/04/13 07:59:17 $
 * $Revision: 1.10 $
 *****************************************************************************/

#include "DefaultBehaviour.hh"

#include "Log.hh"	 // for logging
#include "Timing.hh"
#include "world.hh"
#include "PotentialField.hh"

#include "client.h"	 // for TURNKICKCOMMAND, Vector, AngleDeg, Ray, PlayerObject, Unum, ...
#include "mybehave.h"	 // for face_neck_to_ball()
#include "kick.h"	 // for dokick()


bool DefaultBehaviour::checkAction(Action * action) {

  //  bool ret = isSafeQuicklyChecked( action );
  bool ret = isSafe( action );

  return ret;
}



/** Calculates the predicted velocity which the ball would get if we would kick it to
    the given target point. */
cmu::Vector DefaultBehaviour::getPredictedBallVelocity(cmu::Vector & target) {
  cmu::AngleDeg abs_dir = (target-cmu::Mem->MyPos()).dir();
  cmu::Bool CanKickToTraj;
  cmu::TurnKickCommand tkc = cmu::dokick(abs_dir - cmu::Mem->MyBodyAng(),
					 2*cmu::Mem->SP_ball_speed_max, 1.0, &CanKickToTraj );
  return cmu::Polar2Vector( cmu::Mem->BallKickRate()*tkc.power, abs_dir );
}



/** Bewertet einen Zielpunkt, ob ein Schuss dahin sicher wre. Dabei werden
    Kreise um den Gegner gezogen, und geprft ob die Linie zum Zielpunkt den Kreis
    schneidet. Diese Berechnung ist natrlich ungenau. Aber sie wird wesentlich
    schneller berechnet als eine vollstndige Interception Point Berechnung. */
bool DefaultBehaviour::isSafeFromOpponent(cmu::Vector target, cmu::PlayerObject * opponent) {

  cmu::AngleDeg   angleToTarget     = (target - cmu::Mem->MyPos()).dir();
  cmu::Vector	  opponentPosition  = opponent->get_abs_pos();
  float 	  distanceToTarget  = cmu::Mem->MyPos().dist(target)+2;
  float 	  distanceToOpponent= cmu::Mem->MyPos().dist(opponentPosition);
  cmu::Vector	  sol1,sol2;
  float 	  radius	    = 1.75; // 0.75

  if (distanceToTarget < distanceToOpponent) return true;

  if (distanceToOpponent > 8.0) {
    radius = 4.0;

    return cmu::RayCircleIntersect( cmu::Ray( cmu::Mem->MyPos(), angleToTarget ),
				    radius, opponentPosition,
				    &sol1, &sol2 ) < 2;
  }
  radius = 2.0;

  return cmu::RayCircleIntersect( cmu::Ray( cmu::Mem->MyPos(), angleToTarget ),
				  radius, opponentPosition,
				  &sol1, &sol2 ) < 2;
}


/** Eine schnelle ungenaue Methode. Sie geht alle Gegner durch und schaut, ob
    einer in der Nhe des Weges zum Zielpunkt steht. Gibt nur true zurck, wenn
    alle einen gewissen Abstand zum Weg des Balles haben. Beachtet keine Mitspieler!
    */
bool DefaultBehaviour::isSafeQuicklyChecked(Action * action) {

  int i = 0;

  while (i < cmu::Mem->NumOpponents()) {

    if (i < cmu::Mem->NumOpponents()) {
      cmu::PlayerObject * tempPlayer = cmu::Mem->Opponents()[i];

      if ((tempPlayer != NULL) &&
	  (tempPlayer->pos_valid())) {


	if (!isSafeFromOpponent( cmu::Vector( action->x, action->y ), tempPlayer ) ) {
	  return false;
	}

      }
    }

    i++;
  }

  return true; // checked all opponents, no danger found.
}

/** Bewertet eine Schuss-Aktion, ob sie sicher ist. Sicher heisst, dass der Ball
    nicht abgefangen werden kann. Dabei wird geprft, welcher Spieler am schnellsten
    am Ball sein kann. Dafr muss in das CMU Weltmodell eingegriffen werden, in der
    Art dass die Ballgeschwindigkeit und -richtung (also die Velocity), die vor dem
    Schuss natrlich noch 0 sind, auf einen geschtzten Wert gesetzt werden, bevor
    die Interception Point Berechnung von CMU luft.*/
bool DefaultBehaviour::isSafe(Action * action) {
  //

  cmu::Vector endPos(action->x,action->y);

  cmu::Vector startVelocity = getPredictedBallVelocity( endPos );

  cmu::Vector gPos = cmu::Mem->GetBall()->get_abs_pos();
  gPos += startVelocity;


  // fake Ball
  cmu::Mem->GetBall()->fakeObject( gPos,startVelocity );
  forceAllIPCalculations();


  cmu::Unum fastestTeammate = cmu::Mem->FastestTeammateToBall();
  cmu::Unum fastestOpponent = cmu::Mem->FastestOpponentToBall();

  if (( fastestTeammate == Unum_Unknown )&&( fastestOpponent == Unum_Unknown )) return true;
  if ( fastestTeammate == Unum_Unknown ) return false;
  if ( fastestOpponent == Unum_Unknown ) return true;

  bool retValue = (   cmu::Mem->TeammateInterceptionNumberCycles ( fastestTeammate )     <
		    ( cmu::Mem->OpponentInterceptionNumberCycles ( fastestOpponent )+5 ) );


  // unfake Ball
  cmu::Mem->GetBall()->unfakeObject();
  forceAllIPCalculations();


  return retValue;
}


/** Zweck siehe isSafe. Implementation hnlich, nur anders. */
bool DefaultBehaviour::isSafe2(Action * action) {
  /*
    Was soll diese Methode tun ?

    - Alle Mit- und Gegenspieler ueberpruefen, wann sie fruehestens
      an den Ball kommen
    - Wenn derjenige, der zuerst am Ball sein duerfte, ein Mitspieler ist,
      gebe true zurueck. Ansonsten gebe false zurueck.
  */

  PlayerInterceptionInfo bestPII;
  PlayerInterceptionInfo tempPII;

  int i=0;
  cmu::PlayerObject * tempPlayer = NULL;

  while ((i < cmu::Mem->NumTeammates())&&(i < cmu::Mem->NumOpponents())) {

    if (i < cmu::Mem->NumTeammates()) {
      tempPlayer = cmu::Mem->Teammates()[i];

      if ((tempPlayer != NULL) &&
	  (tempPlayer->pos_valid()) &&
	  (cmu::Mem->MyNumber != tempPlayer->unum)) {
	tempPII = calculateInterceptionInfoForPlayer( tempPlayer,
						      action ) ;
	if (!tempPII.isUnableToIntercept)
	  if (tempPII.distance < bestPII.distance)
	    bestPII = tempPII;
      }
    }
    if (i < cmu::Mem->NumOpponents()) {
      tempPlayer = cmu::Mem->Opponents()[i];

      if ((tempPlayer != NULL) &&
	  (tempPlayer->pos_valid())) {
	tempPII = calculateInterceptionInfoForPlayer( tempPlayer,
						      action ) ;
	if (!tempPII.isUnableToIntercept)
	  if (tempPII.distance < bestPII.distance)
	    bestPII = tempPII;
      }
    }

    i++;
  }

  if ((!bestPII.isUnableToIntercept)&&(bestPII.side == cmu::Mem->TheirSide))
    return false;
  else return true;

}

Action * DefaultBehaviour::selectPossibleAction(list<Action*> * actions) {
  list<Action*>::iterator endIterator = actions->end();
  int c = 0;

  for (list<Action*>::iterator iter = actions->begin(); iter!=endIterator; iter++)
    {
      c++;
      
      Action * tempActionObj = *iter;

      if (tempActionObj->type != ACTION_kick) return tempActionObj;

      if (checkAction(tempActionObj)) {
	Log::debug << "rejected " << c << " actions" << endl;
	return tempActionObj;
      }
      
    }

  // mhhhh.... kein passendes Action-Objekt gefunden
  return NULL;
}



Action * DefaultBehaviour::getAction() {
  //
  static int lastActionCount = 0;

  if (timing->allowToGenerate()) {
    lastActionCount = 0;

    // Hole Infos aus dem Weltmodell, bevor die eigentliche Bewertung
    // ausgefuehrt wird.
    cmu::Mem->getWorld()->update(false);

    if (!list_mode) {

      int * action = cmu::Mem->potentialField()->getEvaluatedAction();

      if (action == NULL) {
	currentAction = NULL;
      } else {

	currentAction = new Action(action);
	timing->registerGeneration();

      }
    } else {

      list<Action*> * actions = cmu::Mem->potentialField()->getListOfEvaluatedActions();

      if (actions == NULL) {
	currentAction = NULL;
      } else {
	Action * tmp = selectPossibleAction( actions );
	if (tmp == NULL) currentAction = NULL;
	else {
	  currentAction = tmp;
	  timing->registerGeneration();
	}
      }
    }


    return currentAction;
  } else {
    lastActionCount++;

    return currentAction;
  }

  return NULL;
}






bool DefaultBehaviour::behave() {

  Action * action = getAction();

  if (action != NULL) {

      if (action->type != ACTION_unknown) {
	  
	  if ( ( action->pf_x == 0 ) &&
	       ( action->pf_y == 0 ) ) {
// Sonderfall 0,0-Aktion Abfangmaneuver

	      cmu::Mem->setInitialValues(cmu::Mem->MyNumber, true);
	      
	  } else {

    switch (action->type) {
    case ACTION_unknown: break;
    case ACTION_run: cmu::Mem->doRun(*action);
      if (debug)
	Log::debug << "action: doRun" << endl;
      break;
    case ACTION_dribble: cmu::Mem->doDribble(*action);
      if (debug)
	Log::debug << "action: doDribble" << endl;
      break;
    case ACTION_getball: cmu::Mem->doGetball(*action);
      if (debug)
	Log::debug << "action: doGetball" << endl;
      break;
    case ACTION_kick: cmu::Mem->doKick(*action);
      if (debug)
	Log::debug << "action: doKick" << endl;
      break;
    case ACTION_move: cmu::Mem->doMove(*action) ;
      if (debug)
	Log::debug << "action: doMove" << endl;
      break;
    }
	  }
      }
  } else {
    if (debug)
      Log::debug << "action: keine!" << endl;
    cmu::face_neck_to_ball();
  }

  return true; // this is the default behaviour which always behaves in the default way.

}



PlayerInterceptionInfo DefaultBehaviour::calculateInterceptionInfoForPlayer( void * player,
									     Action * action ) {
  cmu::PlayerObject * pObject = (cmu::PlayerObject *)player;

  // Diese Methode berechnet einen Interception Point und gibt zurueck, wie weit
  // dieser entfernt ist.

  cmu::Vector beginPos = cmu::Mem->MyPos();
  cmu::Vector endPos(action->x,action->y);

  cmu::Vector startVelocity = (endPos-beginPos).SetLength( cmu::Mem->SP_ball_speed_max * 0.75 );

  // fake Ball
  cmu::Mem->GetBall()->fakeObject( startVelocity );

  // force recalculation of interception points
  cmu::Mem->forceIPCalculationForPlayer(pObject->side,pObject->unum);

  if (pObject->unum == Unum_Unknown)
    return PlayerInterceptionInfo(); // leere info mit unable to intercept auf true

  if (! cmu::Mem->PlayerInterceptionAble(pObject->side,pObject->unum) )
    return PlayerInterceptionInfo(); // leere info mit unable to intercept auf true

  // info mit unable to intercept auf false
  PlayerInterceptionInfo pii( pObject->side,
			      cmu::Mem->PlayerInterceptionPoint(pObject->side,pObject->unum).dist(beginPos) );

  // unfake Ball
  cmu::Mem->GetBall()->unfakeObject();
  // make sure that interception points are recalculated if needed
  cmu::Mem->forceIPCalculationForPlayer(pObject->side,pObject->unum);

  return pii;
}

void DefaultBehaviour::forceAllIPCalculations() {
  cmu::Mem->reinitializeStoredIPs();
}


/*
 * $Log: DefaultBehaviour.cc,v $
 * Revision 1.10  2002/04/13 07:59:17  jens
 * German Open 2002 Edition
 *
 * Revision 1.9  2002/04/08 13:32:09  fastjack
 * listMode jetzt als Property von DefaultBehaviour. per default auf false, Torwart setzt auf true.
 *
 * Revision 1.8  2002/04/08 13:28:06  seekamp
 * Aktionsausfuehrung fuer Manndecker wieder Staminschonender implementiert
 *
 * Revision 1.7  2002/04/05 16:16:28  fastjack
 * kleinere Optimierungen
 *
 * Revision 1.6  2002/04/04 21:15:43  fastjack
 * removed PotentialField code from BTeamAction, all is in DefaultBehaviour now. I hope it works.
 *
 * Revision 1.5  2002/04/04 20:44:04  fastjack
 * bug fix
 *
 * Revision 1.4  2002/04/04 20:37:44  fastjack
 * bug fix
 *
 * Revision 1.3  2002/04/04 20:23:25  fastjack
 * bug fix
 *
 * Revision 1.2  2002/04/04 20:16:18  fastjack
 * moved PotentialField Behaviour in DefaultBehaviour class
 *
 * Revision 1.1  2002/04/02 15:58:11  fastjack
 * Various changes. Torwart Abschlag Bugfix. kleinere unbedeutendere Korrekturen.
 *
 */
