/*****************************************************************************
 * $RCSfile: Timing.cc,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 * 
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: jens $
 * $Date: 2002/04/13 07:59:17 $
 * $Revision: 1.31 $
 *
 *****************************************************************************/


#include "Timing.hh"
#include "client.h"
#include "Log.hh"
#include "PFieldLogger.hh"
#include "PotentialField.hh"

#define INTERVAL_MID 4
#define INTERVAL_LOW 8

void Timing::printDebugInfo() {
  Log::debug << "Timing::blockedCycles = " << blockedCycles << endl;
  Log::debug << "Timing::didGetBall = " << (didGetBall?"true":"false") << endl;
  Log::debug << "Timing::kicksWhileBlockedState = " << kicksWhileBlockedState << endl;
}

void Timing::registerGeneration()  {

  setNextEvalTime( 0 );
  //  setNextEvalTime( (actionArray[4]==0)?0:(cmu::Mem->CurrentTime.t + actionArray[4]) );
  // actionArray[4] == 0 immer :-)

  // Logging Mechanismus fuer Potentialfelder aufrufen.
  PFieldLogger::getInstance().log(cmu::Mem->potentialField());

}

bool Timing::getSeen() {
  //
  bool tmp = seen;
  seen = false;
  return tmp;

}

void Timing::setSeen(bool b) {
  seen = b;
}


bool Timing::allowToGenerate() {
  // Log::info <<"Meine Position: (" <<cmu::Mem->MyX() <<"/" <<cmu::Mem->MyY() <<")" <<endl;
  //da ein harter Schuss aus einer Abfolge von Aktionen besteht, fuehren
  //wir eine Bewertung nur dann aus, wenn wir nicht in einem dieser Schritte
  //sind!!!
  // if  (cmu::Mem->HKStep >-1) {
  //    Log::debug << "Timing::allowToGenerate(): HKStep >-1 -> false" << endl;
  //    return false;
  //  }

  // Log::info << "Meine Stamina: (" << cmu::Mem->MyStamina() <<")" <<endl;

  if (didGetBall) {
    didGetBall = false;
    return true;
  }

  //
  if (blocked() || !getSeen()) return false;
  // else return true;

  // Bei hoher Prioritaet soll die Bewertung IMMER
  // durchgefuehrt werden.
  if (isHighPriority()) {
    nextEvalTime = 0;
    return true;
  }

  // Niedrige Prioritaet
  if (isLowPriority()) {


    /*
      Wenn kein Bewertungszeitpunkt gewuenscht ist,
      dann setzen wir selber eine. Der Agent soll seine
      Bewertungsfunktion ca. alle 2 geraden Zeitpunkte
      ausfuehren.
     */

    if (nextEvalTime == 0)
      nextEvalTime = cmu::Mem->CurrentTime.t + (4-(cmu::Mem->CurrentTime.t % 4));

    /*
      Zeitpunkt ist gesetzt - falls der Zeitpunkt dummerweise
      auf einem Zeitpunkt gelandet ist, der NICHT durch 4
      teilbar ist, muessen wir ihn korrigieren, indem wir das
      Fehlende addieren.
     */
    if (nextEvalTime % 4 != 0) {
      nextEvalTime += (4-(nextEvalTime % 4));
      // Log::info << "Timing: (low priority) do not generate pf (nextEvalTime:" << nextEvalTime << ")" << endl;
      return false;

    } else {

      /* Bewertung ausfuehren, wenn die Zeit gekommen ist. */
      if (cmu::Mem->CurrentTime.t >= nextEvalTime) {
	nextEvalTime = 0;
	// Log::info << "Timing: (low priority) DO generate pf (nextEvalTime:" << nextEvalTime << ")" << endl;
	return true;
      }
      else {
	// Log::info << "Timing: (low priority) do not generate pf (nextEvalTime:" << nextEvalTime << ")" << endl;
	return false;
      }
    }
  }


  // Mittlere Prioritaet
  if (isMidPriority()) {

    /*
      Wenn kein Bewertungszeitpunkt gewuenscht ist,
      dann setzen wir selber eine. Der Agent soll seine
      Bewertungsfunktion ca. alle ungeraden Zeitpunkte
      ausfuehren.
     */

    if (nextEvalTime == 0)
      nextEvalTime = cmu::Mem->CurrentTime.t + ((cmu::Mem->CurrentTime.t%2 != 0) ? 2 : 3);

    /*
      Zeitpunkt ist gesetzt - falls der Zeitpunkt dummerweise
      auf einem geraden Zeitpunkt gelandet ist, wird er
      korrigiert, indem er inkrementiert wird.
     */
    if (nextEvalTime % 2 == 0) {
      nextEvalTime++;
      // Log::info << "Timing (mid priority): do not generate pf (nextEvalTime:" << nextEvalTime << ")" << endl;
      return false;

    } else {

      /* Bewertung ausfuehren, wenn die Zeit gekommen ist. */
      if (cmu::Mem->CurrentTime.t >= nextEvalTime) {
	nextEvalTime = 0;
	// Log::info << "Timing: (mid priority) DO generate pf (nextEvalTime:" << nextEvalTime << ")" << endl;
	return true;
      }
      else {
	// Log::info << "Timing (mid priority): do not generate pf (nextEvalTime:" << nextEvalTime << ")" << endl;
	return false;
      }


    }
  }
  return false;
}

bool Timing::blocked() {
  if (blockedCycles > 0) {
    blockedCycles--;
    return true;
  } else {
    kicksWhileBlockedState = 0;
    return false;
  }
}

bool Timing::isHighPriority()
{
  /* test: wir sehen mal ob wir mit immer hoher prioritaet klar kommen */
  // return true;

    if ((cmu::Mem->MyNumber==1) && (cmu::Mem->MyX()<-25))
	return true;
    
    if ((cmu::Mem->BallAbsolutePosition().x<-25) && (cmu::Mem->MyNumber<7))
	return true;
    
    float distance = cmu::Mem->DistanceTo(cmu::Mem->BallAbsolutePosition());
    return (distance <= NEAR_TO_BALL);
}

bool Timing::isMidPriority() {
  float distance = cmu::Mem->DistanceTo(cmu::Mem->BallAbsolutePosition());
  return (distance <= MID_TO_BALL);
}

bool Timing::isLowPriority() {
  return !isMidPriority();
}



/*
 * $Log: Timing.cc,v $
 * Revision 1.31  2002/04/13 07:59:17  jens
 * German Open 2002 Edition
 *
 * Revision 1.30  2002/04/05 08:40:57  seekamp
 * Wenn Ball in eigenem Viertel, dann sollen alle Aktionen der Abwehrspieler ausgefuehrt werden
 *
 * Revision 1.29  2002/03/05 17:08:01  fastjack
 * Interception Methoden fast fertig
 *
 * Revision 1.28  2002/03/04 16:02:14  fastjack
 * Interception Points angefangen zu implementieren. ACHTUNG code nicht lauffaehig
 *
 * Revision 1.27  2002/03/01 11:18:47  fastjack
 * GPL added
 *
 * Revision 1.26  2002/02/26 17:59:30  jens
 * weiter am parsen eines eigenen PF gearbeitet. Einbinden in das Weltmodell funzt noch nicht richtig
 *
 * Revision 1.25  2002/02/15 17:33:52  seekamp
 * neuen cone getestet - verbessertes Abwehrverhalten
 *
 * Revision 1.24  2002/02/10 12:19:14  fastjack
 * Habe die Includes angepasst, damit die Header sich so wenig wie mglich
 * gegenseitig includen. Z.T. ersetzt durch forward Deklaration. getPFEntry Methode
 * aus dem PotentialField heraus geholt und in den PFieldLogger gepackt - ist dort
 * sinniger.
 *
 * Revision 1.23  2002/02/07 20:07:51  fastjack
 * merged
 *
 * Revision 1.22  2002/02/05 19:14:01  fastjack
 * Versucht, den Interception-Reflex zu optimieren.
 *
 * Revision 1.21  2002/01/29 12:06:38  fastjack
 * debugging in progress
 *
 * Revision 1.20  2002/01/29 11:13:34  fastjack
 * Timing update
 *
 * Revision 1.19  2002/01/29 10:57:59  jens
 * kick verbessert und pausen der Spieler teilweise behoben (scan field with body eingesetzt)
 *
 * Revision 1.18  2002/01/25 18:48:32  seekamp
 * Staminaverwaltung implementiert
 *
 * Revision 1.17  2002/01/22 18:05:41  jens
 * Some Staminaproblems solved
 *
 * Revision 1.16  2002/01/22 14:07:50  fastjack
 * Timing verbessert
 *
 * Revision 1.15  2002/01/22 12:06:13  fastjack
 * Timing tests
 *
 * Revision 1.14  2002/01/21 17:40:44  seekamp
 * dynamisch anpassbare Rastergroesse implementiert
 *
 * Revision 1.13  2002/01/18 20:11:21  fastjack
 * Das Logging von Potentialfeldern fertig gestellt. Das Logging ist auch
 * ein-/ausschaltbar ( PFieldLogger::getInstance().setEnabled(true) ).
 *
 * Revision 1.12  2002/01/17 19:11:40  fastjack
 * kleinere Aenderungen bzgl. pfield logging eingefuegt. dummy methode in PotentialField eingefuegt ...
 *
 * Revision 1.11  2002/01/17 18:15:30  seekamp
 * Verschiedene Methoden in Potentialfield und BTeamAction veraendert, mit dem
 * Ergebnis, dass der ballnaechste Spieler zum Ball laeuft und einen Pass spielt.
 * (mehr oder weniger)
 *
 * Revision 1.10  2002/01/15 21:50:37  seekamp
 * Fehler in der Bewertungsfunktion ausgemerzt.
 * Individuelles Spielverhalten getestet.
 *
 * Revision 1.9  2002/01/15 16:35:04  jens
 * nur ein kleines update bzgl. dem PFieldLogger
 *
 * Revision 1.8  2002/01/10 19:02:45  jens
 * tries to find further bugs. The actual bugs are: the Ballposition seems to be invalid sometimes. We dont know where the ballposition becomes set and why we have an invalid position although we have a valid position IN the Worldmodel of CMU. The second bug we tried to fix is an invalid position of player. It seems that only the last player has an invalid position.
 *
 * Revision 1.7  2002/01/08 17:20:49  jens
 * eliminated two major bugs which caused the clients to exit. Did it with the use of some small hacks and a lot of bugfinding. Three Flags(two for the close goals and one for a flag which CMU doesnt understands) where omitted, but can be reintegrated easily(Visual Handler).
 *
 * Revision 1.6  2002/01/07 18:50:11  jens
 * Further tryings to get rid of those Bugs. Found two sources, eliminated one
 * Actual Bug is located in the cenvesrion from RCC-Flags to CMU-Markers.
 *
 * Revision 1.5  2002/01/03 01:04:38  fastjack
 * Parser verwendet fuer hear Messages wieder die originale CMU Methode. Ausserdem habe ich einige weitere Ausgaben in das PotentialField getan, um herauszufinden, wo er aussteigt. Dann habe ich das Timing so geHACKT, dass es immer true gibt, ausser es ist geblockt. Debugging necessary.
 *
 * Revision 1.4  2001/12/28 16:10:33  fastjack
 * verschiedenste bugfixes in allen bereichen
 *
 * Revision 1.3  2001/12/20 18:16:24  jens
 * eleminated different compile errors and started to implement simple actions
 *
 * Revision 1.2  2001/12/14 20:41:40  fastjack
 * PotentialField neu geschrieben auf Basis von Matthias' Code
 *
 * Revision 1.1  2001/12/13 20:39:01  fastjack
 * Timing Klasse entwickelt. Ist noch Alpha :-)
 *
 */
