/*****************************************************************************
 * $RCSfile: gouraud.hh,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 * 
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: fastjack $
 * $Date: 2002/03/18 17:13:03 $
 * $Revision: 1.7 $
 *****************************************************************************/

#ifndef GOURAUDFILLER_HH
#define GOURAUDFILLER_HH

enum FillMode {
  FM_ADD,
  FM_AND,
  FM_OR,
  FM_XOR,
  FM_MULTIPLY,
  FM_REPLACE
};


class GouraudFiller {

  int *field;
  int dimx;
  int dimy;
  int fieldsize;
  float cost[360];
  float sint[360];

  int *ledge;
  int *redge;
  int *lcolor;
  int *rcolor;
  
  FillMode fillMode;


  bool initializedSinCos;
  void buildCosineTable();
  void buildSineTable();
  void rotatePointAroundPoint(int &x,int &y,
                              int px,int py,
                              int degree);
  void initSinCosTables() {
    if (!initializedSinCos) {
      buildCosineTable(); buildSineTable();
      initializedSinCos=true;
    }
  }

  int getValue(int x, int y);
  void setValue(int v, int x, int y);

  void fillHLine(int x1, int x2, int y, int v1, int v2);

  void scanEdge(int *edge, int *color,
		int x1, int x2, int y1, int y2, int c1, int c2);

  void gouraudFillTriangle(int x1,int x2,int x3,
                           int y1,int y2,int y3,
                           int c1,int c2,int c3);
  void sortValues(int &x1, int &y1, int &c1,
		  int &x2, int &y2, int &c2,
		  int &x3, int &y3, int &c3);

  inline int int_abs(int a);
  inline float int_pow(float x, int p);
  inline float round(float x, int p=0);
  inline bool isInBounds(int x, int y);
  void clearArrays(int from=0, int len=-1);

public:

  GouraudFiller(int * f, int dx, int dy);
  virtual ~GouraudFiller();

  void setFillMode(FillMode fm) { fillMode = fm; }
  FillMode getFillMode() { return fillMode; }

  void shadeTriangle( int x1,int y1, int color1,
                      int x2,int y2, int color2,
                      int x3,int y3, int color3 );

  void shadePyramid( int midx,int midy, int halfSideLength, int height );

    /**
     * Will become two triangles only.
     * Has four vertexes - starting on the lower left, going CCW.
     * Like this:<br />
     * <pre>
     *    D +---------------+ C
     *       \      |h     /
     *         \    |    /
     *           \+-P-+/
     *          A       B
     * </pre>
     * Where A to D are the four vertexes, and P should be the point
     * given by the first two parameters - being x and y.<br />
     *
     * The length of the straight line AB is give with widthNear,
     * while widthFar gives the length of the straight line DC. The
     * length of the vertical line with the little h is given by the
     * parameter length.<br />
     * Last but not least: the parameter orientation tells us how to
     * rotate the trapezoid around the point P. 0 would be &quot;not
     * rotate&quot;, while 2&pi; would be equal to 360 degrees and thus
     * also the same as 0.
     */
  void shadeTrapezoid( int x, int y,
                       int widthNear, int widthFar,
                       int length, float orientation,
                       int h1,int h2,int h3,int h4);

};

#endif

/*
 * $Log: gouraud.hh,v $
 * Revision 1.7  2002/03/18 17:13:03  fastjack
 * Absturzursache im GouraudFiller behoben.
 *
 * Revision 1.6  2002/03/15 13:26:01  fastjack
 * Ausgaben entfernt / dem Gouraud Kerl mehr Genauigkeit verschafft und einen boesen BUG entfernt
 *
 * Revision 1.5  2002/03/01 11:18:48  fastjack
 * GPL added
 *
 * Revision 1.4  2002/02/15 13:51:10  fastjack
 * GouraudFiller repariert, fehlerhaften Algorithmus gefixt. Paar Kommentare entfernt.
 *
 * Revision 1.3  2002/02/14 19:39:54  fastjack
 * Gouraud Shader verbessert - sind noch Probleme drin leider.
 *
 * Revision 1.2  2002/02/12 12:09:18  fastjack
 * Alternative Cone Methode ZUM TESTEN eingefuegt. Bei Nichtgefallen wieder entfernen :-D
 *
 * Revision 1.1  2002/02/12 00:52:01  fastjack
 * new files
 *
 */

